try:
    from IPython import get_ipython

    ipy = get_ipython()
    if ipy is not None:
        shell = ipy.__class__.__name__
        if shell == "ZMQInteractiveShell":
            IN_JUPYTER = True
        else:
            IN_JUPYTER = False
    else:
        IN_JUPYTER = False
except (NameError, ImportError, AttributeError):
    IN_JUPYTER = False



import hashlib
import os
from pathlib import Path


def dir_hash(directory: str):
    md5 = hashlib.md5()
    for root, dirs, files in os.walk(directory):
        for file in files:
            path = os.path.join(root, file)
            modify_time = os.path.getmtime(path)
            md5.update(f"{path}{modify_time}".encode())
    return md5.hexdigest()


def ensure_dir(directory: str):
    Path(directory).mkdir(parents=True, exist_ok=True)



import hashlib
import numpy as np
import os
import pandas as pd
import re
from typing import Union

from .constants import (
    COL_COUNT,
    COL_FILE_NAME,
    COL_PROC_NAME,
    COL_SIZE,
    COL_TIME,
    COL_TIME_RANGE,
    COL_TIME_START,
    FILE_PATTERN_PLACEHOLDER,
    PROC_NAME_SEPARATOR,
    SIZE_BINS,
    SIZE_BIN_SUFFIXES,
)


def fix_dtypes(df: pd.DataFrame):
    int_cols = []
    int_cols.extend([col for col in df.columns if '_bin_' in col])
    int_cols.extend([col for col in df.columns if col.endswith('_count')])
    int_cols.extend([col for col in df.columns if col.endswith('_nunique')])
    double_cols = []
    double_cols.extend([col for col in df.columns if col.endswith('_bw')])
    double_cols.extend([col for col in df.columns if col.endswith('_intensity')])
    double_cols.extend([col for col in df.columns if col.endswith('_ops')])
    double_cols.extend([col for col in df.columns if col.endswith('_time')])
    double_cols.extend([col for col in df.columns if col.endswith('_slope')])
    double_cols.extend([col for col in df.columns if col.endswith('_pct')])
    size_cols = [col for col in df.columns if col.endswith('_size')]
    df[int_cols] = df[int_cols].astype('Int64')
    df[double_cols] = df[double_cols].astype('Float64')
    df[size_cols] = df[size_cols].astype('Int64')
    return df


def fix_size_values(df: pd.DataFrame):
    size_cols = [col for col in df.columns if 'size' in col]
    df[size_cols] = df[size_cols].replace(0, np.nan)
    return df


def set_app_name(df: pd.DataFrame):
    return df.assign(
        app_name=lambda df: df.index.get_level_values(COL_PROC_NAME)
        .str.split(PROC_NAME_SEPARATOR)
        .str[0]
        .astype("string[pyarrow]"),
    )


def set_host_name(df: pd.DataFrame):
    return df.assign(
        app_name=lambda df: df.index.get_level_values(COL_PROC_NAME)
        .str.split(PROC_NAME_SEPARATOR)
        .str[1]
        .astype("string[pyarrow]"),
    )


def set_file_dir(df: pd.DataFrame):
    if COL_FILE_NAME not in df.index.names:
        return df
    return df.assign(
        file_dir=df.index.get_level_values(COL_FILE_NAME).map(os.path.dirname).astype("string[pyarrow]"),
    )


def set_file_pattern(df: pd.DataFrame):
    if COL_FILE_NAME not in df.index.names:
        return df

    def _apply_regex(file_name: str):
        return re.sub('[0-9]+', FILE_PATTERN_PLACEHOLDER, file_name)

    return df.assign(
        file_pattern=df.index.get_level_values(COL_FILE_NAME).map(_apply_regex).astype("string[pyarrow]"),
    )


def set_id(ix: Union[tuple, str, int]):
    ix_str = '_'.join(map(str, ix)) if isinstance(ix, tuple) else str(ix)
    return int(hashlib.md5(ix_str.encode()).hexdigest(), 16)


def set_proc_name_parts(df: pd.DataFrame):
    if COL_PROC_NAME not in df.index.names:
        return df

    proc_names = df.index.get_level_values(COL_PROC_NAME)

    first_proc_name_parts = proc_names[0].split(PROC_NAME_SEPARATOR)

    if first_proc_name_parts[0] == 'app':
        return df.assign(
            proc_name_parts=proc_names.str.split(PROC_NAME_SEPARATOR),
            app_name=lambda df: df.proc_name_parts.str[0].astype(str),
            host_name=lambda df: df.proc_name_parts.str[1].astype(str),
            # node_name=lambda df: pd.NA,
            proc_id=lambda df: df.proc_name_parts.str[2].astype(str),
            rank=lambda df: pd.NA,
            thread_id=lambda df: df.proc_name_parts.str[3].astype(str),
        ).drop(columns=['proc_name_parts'])

    return df.assign(
        proc_name_parts=proc_names.str.split(PROC_NAME_SEPARATOR),
        app_name=lambda df: df.proc_name_parts.str[0].astype(str),
        host_name=lambda df: df.proc_name_parts.str[1].astype(str),
        # node_name=lambda df: df.proc_name_parts.str[1].astype(str),
        proc_id=lambda df: pd.NA,
        rank=lambda df: df.proc_name_parts.str[2].astype(str),
        thread_id=lambda df: pd.NA,
    ).drop(columns=['proc_name_parts'])


def set_size_bins(df: pd.DataFrame):
    df['size_bin_temp'] = pd.cut(
        df['size'],
        bins=SIZE_BINS,
        labels=SIZE_BIN_SUFFIXES,
        right=True,
        include_lowest=True,
    )
    size_bin_dummies = pd.get_dummies(df['size_bin_temp'], prefix='size_bin', dtype=int)
    df = pd.concat([df, size_bin_dummies], axis=1)
    df = df.drop('size_bin_temp', axis=1)
    return df


def set_unique_counts(df: pd.DataFrame, layer: str):
    unique_cols = [col for col in df.columns if col.endswith('_unique')]
    for unique_col in unique_cols:
        if COL_FILE_NAME in unique_col and 'posix' not in layer:
            continue
        nunique_col = unique_col.replace('_unique', '_nunique')
        # Handle null values before applying len() - null values should map to 0
        df[nunique_col] = df[unique_col].map(lambda x: len(x) if pd.notna(x) else 0).astype('uint64[pyarrow]')
    return df.drop(columns=unique_cols)


def split_duration_records_vectorized(
    df: pd.DataFrame,
    time_granularity: float,
    time_resolution: float,
) -> pd.DataFrame:
    # Convert duration column to numpy array
    durations = df[COL_TIME].to_numpy()

    # Calculate number of chunks needed for each row
    n_chunks = np.ceil(durations / time_granularity).astype(int)
    max_chunks = n_chunks.max()

    if max_chunks == 0:
        df[COL_TIME_RANGE] = df[COL_TIME_START] // (time_granularity * time_resolution)
        df[COL_TIME_RANGE] = df[COL_TIME_RANGE].astype('int64')
        return df.copy()

    # Create expansion indices
    row_idx = np.arange(len(df))
    repeated_idx = np.repeat(row_idx, n_chunks)

    # Create the expanded dataframe
    result_df = df.iloc[repeated_idx].copy()

    # Calculate chunk numbers for each expanded row
    chunk_numbers = np.concatenate([np.arange(n) for n in n_chunks])

    # Calculate time values
    is_last_chunk = chunk_numbers == (n_chunks.repeat(n_chunks) - 1)
    time_values = np.full(len(chunk_numbers), time_granularity)

    # Handle remainders for last chunks
    remainders = durations % time_granularity
    remainders = np.where(remainders == 0, time_granularity, remainders)
    time_values[is_last_chunk] = remainders.repeat(n_chunks)[is_last_chunk]

    # Update time column
    result_df[COL_TIME] = time_values

    # Calculate and update timestamps
    ts_base = df[COL_TIME_START].to_numpy()
    ts_offsets = chunk_numbers * time_granularity * time_resolution  # Convert to microseconds
    result_df[COL_TIME_START] = ts_base.repeat(n_chunks) + ts_offsets

    result_df[COL_TIME_RANGE] = result_df[COL_TIME_START] // (time_granularity * time_resolution)
    result_df[COL_TIME_RANGE] = result_df[COL_TIME_RANGE].astype('int64')

    counts = df[COL_COUNT].to_numpy()
    expanded_counts = counts.repeat(n_chunks)
    # Divide counts by number of chunks for each original row
    result_df[COL_COUNT] = expanded_counts / n_chunks.repeat(n_chunks)

    sizes = df[COL_SIZE].to_numpy()
    expanded_sizes = sizes.repeat(n_chunks)
    # Divide sizes by number of chunks for each original row
    result_df[COL_SIZE] = expanded_sizes / n_chunks.repeat(n_chunks)

    return result_df.reset_index(drop=True)



import abc
import colorsys
import dask
import dataclasses as dc
import inflect
import pandas as pd
from rich.console import Console
from rich.table import Table
from typing import Dict, List, Optional

from .constants import COL_PROC_NAME, Layer, MiB
from .types import (
    AnalyzerResultType,
    RawStats,
    ViewKey,
    humanized_view_name,
)


@dc.dataclass
class OutputLayerMetrics:
    time: float
    count: int
    size: float
    ops: float
    bandwidth: float
    num_files: int
    num_processes: int
    u_time: Optional[float]
    u_count: Optional[int]
    u_size: Optional[float]


@dc.dataclass
class OutputSummary:
    job_time: float
    layer_metrics: Dict[Layer, OutputLayerMetrics]
    layers: List[Layer]
    time_granularity: float
    time_resolution: float
    total_event_count: int
    unique_file_count: int
    unique_host_count: int
    unique_process_count: int


class Output(abc.ABC):
    def __init__(
        self,
        compact: bool = False,
        name: str = "",
        root_only: bool = False,
        view_names: List[str] = [],
    ):
        self.compact = compact
        self.name = name
        # self.output_dir = HydraConfig.get().runtime.output_dir
        self.pluralize = inflect.engine()
        self.root_only = root_only
        self.view_names = view_names

    def handle_result(self, result: AnalyzerResultType):
        raise NotImplementedError

    def _create_summary(self, result: AnalyzerResultType, view_key: ViewKey) -> OutputSummary:
        flat_view = result.flat_views[view_key]
        raw_stats = dask.compute(result.raw_stats)[0]
        if isinstance(raw_stats, dict):
            raw_stats = RawStats(**raw_stats)
        summary = OutputSummary(
            job_time=float(raw_stats.job_time),
            layer_metrics={},
            layers=result.layers,
            time_granularity=float(raw_stats.time_granularity),
            time_resolution=float(raw_stats.time_resolution),
            total_event_count=int(raw_stats.total_event_count),
            unique_file_count=int(raw_stats.unique_file_count),
            unique_host_count=int(raw_stats.unique_host_count),
            unique_process_count=int(raw_stats.unique_process_count),
        )
        is_process_based = view_key[-1] == COL_PROC_NAME
        time_metric = 'time_sum' if is_process_based else 'time_max'
        for layer in result.layers:
            times = flat_view.get(f"{layer}_{time_metric}", pd.Series([0.0]))
            time = times.max() if is_process_based else times.sum()
            count = flat_view.get(f"{layer}_count_sum", pd.Series([0])).sum()
            size = None
            if 'posix' in layer:
                size = flat_view.get(f"{layer}_size_sum", pd.Series([0.0])).sum()
            num_files = flat_view.get(f"{layer}_file_name_nunique", pd.Series([0.0])).max()
            num_processes = flat_view.get(f"{layer}_proc_name_nunique", pd.Series([0.0])).max()
            u_time_col = f"u_{layer}_{time_metric}"
            u_time = None
            u_count = None
            u_size = None
            if u_time_col in flat_view:
                u_time_mask = flat_view[u_time_col] > 0
                u_times = flat_view.get(u_time_col, pd.Series([0.0]))[u_time_mask]
                u_time = 0.0
                if u_times.any():
                    u_time = u_times.max() if is_process_based else u_times.sum()
                u_counts = flat_view.get(f"{layer}_count_sum", pd.Series([0.0]))[u_time_mask]
                u_count = u_counts.sum() if u_counts.any() else 0
                if 'posix' in layer:
                    u_sizes = flat_view.get(f"{layer}_size_sum", pd.Series([0.0]))[u_time_mask]
                    u_size = u_sizes.sum() if u_sizes.any() else 0.0
            summary.layer_metrics[layer] = OutputLayerMetrics(
                time=float('nan') if pd.isna(time) else float(time),
                count=int(count),
                size=float('nan') if pd.isna(size) else float(size),
                ops=float('nan') if pd.isna(time) or time == 0 else float(count / time),
                bandwidth=float('nan') if pd.isna(time) or pd.isna(size) or time == 0 else float(size / time),
                num_files=0 if pd.isna(num_files) else int(num_files),
                num_processes=0 if pd.isna(num_processes) else int(num_processes),
                u_time=None if pd.isna(u_time) else float(u_time),
                u_count=None if pd.isna(u_count) else int(u_count),
                u_size=None if pd.isna(u_size) else float(u_size),
            )
        return summary

    def _humanized_layer_name(self, name: str) -> str:
        return (
            name.replace('_', ' ')
            .title()
            .replace('Posix', 'POSIX')
            .replace('Stdio', 'STDIO')
            .replace('Lustre', '(Lustre)')
            .replace('Ssd', '(SSD)')
        )


class ConsoleOutput(Output):
    def __init__(
        self,
        compact: bool = False,
        name: str = "",
        root_only: bool = False,
        show_debug: bool = False,
        show_header: bool = True,
        view_names: List[str] = [],
    ):
        super().__init__(compact, name, root_only, view_names)
        self.show_debug = show_debug
        self.show_header = show_header

    def handle_result(self, result: AnalyzerResultType):
        print_objects = []
        for view_key in result.flat_views:
            if view_key[-1] not in result.view_types:
                continue
            summary = self._create_summary(result=result, view_key=view_key)
            summary_table = self._create_summary_table(summary=summary, view_key=view_key)
            layer_breakdown_table = self._create_layer_breakdown_table(summary=summary, view_key=view_key)
            print_objects.append(summary_table)
            print_objects.append(layer_breakdown_table)
        console = Console(record=True)
        console.print(*print_objects)

    def _create_layer_breakdown_table(self, summary: OutputSummary, view_key: ViewKey) -> Table:
        breakdown_table_title = "Layer Breakdown"
        show_overlap = False
        if len(summary.layers) > 1:
            breakdown_table_title += " (w/ overlap %)"
            show_overlap = True
        breakdown_table = Table(title=breakdown_table_title, title_style="bold cyan", expand=True)
        breakdown_table.add_column("Layer", style="bold")
        breakdown_table.add_column("Time (s)", justify="right")
        breakdown_table.add_column("Ops", justify="right")
        breakdown_table.add_column("Ops/sec", justify="right")
        breakdown_table.add_column("Size (MB)", justify="right")
        breakdown_table.add_column("Bandwidth (MB/s)", justify="right")
        for layer in summary.layers:
            layer_metrics = summary.layer_metrics[layer]
            if layer_metrics.count == 0:
                continue
            if show_overlap:
                time_str = self._format_val_with_ovlp_pct(layer_metrics.time, layer_metrics.u_time)
                count_str = self._format_val_with_ovlp_pct(layer_metrics.count, layer_metrics.u_count, fmt_int=True)
            else:
                time_str = self._format_val(layer_metrics.time)
                count_str = self._format_val(layer_metrics.count, fmt_int=True)
            size_str = '-'
            if not pd.isna(layer_metrics.size):
                size_value = layer_metrics.size / MiB
                u_size_value = layer_metrics.u_size / MiB if layer_metrics.u_size is not None else None
                if show_overlap:
                    size_str = self._format_val_with_ovlp_pct(size_value, u_size_value)
                else:
                    size_str = self._format_val(size_value)
            ops_str = f"{layer_metrics.ops:.3f}"
            bandwidth_str = '-'
            if not pd.isna(layer_metrics.bandwidth):
                bandwidth_str = f"{layer_metrics.bandwidth / MiB:.3f}"
            breakdown_table.add_row(
                self._humanized_layer_name(layer),
                time_str,
                count_str,
                ops_str,
                size_str,
                bandwidth_str,
            )
        return breakdown_table

    def _create_summary_table(self, summary: OutputSummary, view_key: ViewKey) -> Table:
        view_name = humanized_view_name(view_key, ' ')

        summary_table = Table(title=f"{view_name} Summary", title_style='bold green', expand=True)
        summary_table.add_column(header='Metric', style='bold')
        summary_table.add_column(header='Unit', style='italic')
        summary_table.add_column(header='Value', justify='right')

        summary_table.add_row('Job Time', 'seconds', f"{summary.job_time:.3f}")
        summary_table.add_row('Total Count', 'count', f"{summary.total_event_count:,}")
        summary_table.add_row('Total Files', 'count', f"{summary.unique_file_count:,}")
        summary_table.add_row('Total Nodes', 'count', f"{summary.unique_host_count:,}")
        summary_table.add_row('Total Processes', 'count', f"{summary.unique_process_count:,}")

        for layer in summary.layers:
            layer_name = self._humanized_layer_name(layer)
            layer_metrics = summary.layer_metrics[layer]
            if layer_metrics.count == 0:
                continue
            summary_table.add_row(f"{layer_name} Count", 'count', f"{layer_metrics.count:,}")
            if layer_metrics.size > 0:
                avg_xfer_size = layer_metrics.size / layer_metrics.count
                summary_table.add_row(f"{layer_name} Size", 'MB', f"{layer_metrics.size / MiB:.3f}")
                summary_table.add_row(f"{layer_name} Bandwidth", 'MB/s', f"{layer_metrics.bandwidth / MiB:.3f}")
                summary_table.add_row(f"{layer_name} Avg Transfer Size", 'MB', f"{avg_xfer_size / MiB:.3f}")

        return summary_table

    def _format_val(self, value: float, fmt_int=False) -> str:
        if value is None or value == 0:
            return '-'
        if fmt_int:
            return f"{int(value):,}"
        return f"{value:.3f}"

    def _format_val_with_ovlp_pct(self, value, u_value, fmt_int=False):
        value = value or 0
        if value == 0 or u_value is None:
            value = self._format_val(value, fmt_int)
            return f"{value} (" + ('-' * 4) + ")"
        u_value = u_value or 0
        ovlp_pct = max(0.0, 1.0 - (u_value / value))
        padded_percent = f"{int(round(ovlp_pct * 100)):>3d}%"
        color = self._percentage_color(ovlp_pct)
        value = self._format_val(value, fmt_int)
        return f"{value} ([{color}]{padded_percent}[/{color}])"

    def _percentage_color(self, percentage: float) -> str:
        """
        Convert overlap percentage (0.0-1.0) to color name in hex (for rich).
        0% = red, 100% = green
        """
        percentage = max(0.0, min(1.0, percentage))  # Clamp to [0, 1]
        # HSV: Hue 0.0 (red) to 0.33 (green)
        h = percentage * 0.33  # red to green
        r, g, b = colorsys.hsv_to_rgb(h, 1.0, 1.0)
        return f"#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}"


class CSVOutput(Output):
    def handle_result(self, result: AnalyzerResultType):
        raise NotImplementedError("CSVOutput is not implemented yet.")


class SQLiteOutput(Output):
    def __init__(
        self,
        compact: bool = False,
        group_behavior: bool = False,
        name: str = "",
        root_only: bool = False,
        run_db_path: str = "",
        view_names: List[str] = [],
    ):
        super().__init__(compact, name, root_only, view_names)
        self.run_db_path = run_db_path

    def handle_result(self, result: AnalyzerResultType):
        raise NotImplementedError("SQLiteOutput is not implemented yet.")



import dask.dataframe as dd
import itertools as it
import numpy as np
import pandas as pd
import portion as P


def nunique():
    return dd.Aggregation(
        name="nunique",
        chunk=lambda s: s.apply(lambda x: list(set(x))),
        agg=lambda s0: s0.obj.groupby(level=list(range(s0.obj.index.nlevels))).sum(),
        finalize=lambda s1: s1.apply(lambda final: len(set(final))),
    )


def quantile_stats(min, max):
    def quantile_stats_finalize(values_list):
        if len(values_list) == 0:
            return [np.nan, np.nan, np.nan]
        values_array = np.array(values_list)
        q_min, q_max = np.quantile(values_array, [min, max])
        filtered_mask = (values_array >= q_min) & (values_array <= q_max)
        filtered_values = values_array[filtered_mask]
        if len(filtered_values) == 0:
            return [np.nan, np.nan, np.nan]
        return [np.mean(filtered_values), np.std(filtered_values), len(filtered_values)]

    return dd.Aggregation(
        f"q{min * 100:.0f}_q{max * 100:.0f}_stats",
        lambda s: s.apply(lambda x: x.replace(0, np.nan).dropna().tolist()),
        lambda s0: s0.obj.groupby(level=0).sum(),
        lambda s1: s1.apply(quantile_stats_finalize),
    )


def unique_set():
    return dd.Aggregation(
        'unique',
        lambda s: s.apply(lambda x: set() if pd.isna(x).any() else set(x)),
        lambda s0: s0.apply(lambda x: set(it.chain.from_iterable(x)) if not pd.isna(x).any() else set()),
        lambda s1: s1.apply(lambda x: set(x) if len(x) > 0 else pd.NA),
    )


def unique_set_flatten():
    def safe_union(x):
        # Handle null values
        if pd.isna(x).any() if hasattr(x, 'any') else pd.isna(x):
            return set()

        # If x is a Series, convert each element to set and union them
        if isinstance(x, pd.Series):
            result = set()
            for item in x:
                if pd.isna(item):
                    continue
                elif hasattr(item, '__iter__') and not isinstance(item, (str, bytes)):
                    try:
                        result.update(item)
                    except TypeError:
                        result.add(item)
                else:
                    result.add(item)
            return result

        # Handle scalar or single iterable
        if hasattr(x, '__iter__') and not isinstance(x, (str, bytes)):
            try:
                return set(x)
            except TypeError:
                return {x}
        else:
            return {x}

    def safe_agg_union(x):
        # Aggregate function to union all sets
        if isinstance(x, pd.Series):
            result = set()
            for item in x:
                if isinstance(item, set):
                    result.update(item)
                elif not pd.isna(item):
                    result.add(item)
            return result
        else:
            return {x} if not pd.isna(x) else set()

    return dd.Aggregation(
        'unique',
        lambda s: s.apply(safe_union),
        lambda s0: s0.agg(safe_agg_union),
        lambda s1: s1.apply(set),
    )


def union_portions():
    def union_s(s):
        emp = P.empty()
        for x in s:
            emp = emp | x
        return emp

    def fin(s):
        val = 0.0
        for i in s:
            if not i.is_empty():
                val += i.upper - i.lower
        return val

    return dd.Aggregation(
        'portion',
        union_s,
        union_s,
        fin,
    )



import numpy as np
from enum import Enum, auto
from strenum import StrEnum


class AccessPattern(Enum):
    SEQUENTIAL = 0
    RANDOM = 1


class EventType(StrEnum):
    ATTACH_REASONS = auto()
    COMPUTE_HLM = auto()
    COMPUTE_MAIN_VIEW = auto()
    COMPUTE_METRIC_BOUNDARIES = auto()
    COMPUTE_PERSPECTIVES = auto()
    COMPUTE_VIEW = auto()
    DETECT_CHARACTERISTICS = auto()
    EVALUATE_VIEW = auto()
    READ_TRACES = auto()
    SAVE_VIEWS = auto()


class IOCategory(Enum):
    READ = 1
    WRITE = 2
    METADATA = 3
    PCTL = 4
    IPC = 5
    OTHER = 6
    SYNC = 7


class Layer(StrEnum):
    APP = auto()
    DATALOADER = auto()
    NETCDF = auto()
    PNETCDF = auto()
    HDF5 = auto()
    MPI = auto()
    POSIX = auto()


COL_ACC_PAT = 'acc_pat'
COL_APP_NAME = 'app_name'
COL_BEHAVIOR = 'behavior'
COL_CATEGORY = 'cat'
COL_COUNT = 'count'
COL_EPOCH = 'epoch'
COL_FILE_DIR = 'file_dir'
COL_FILE_NAME = 'file_name'
COL_FILE_PATTERN = 'file_pattern'
COL_FUNC_NAME = 'func_name'
COL_HOST_NAME = 'host_name'
COL_IO_CAT = 'io_cat'
COL_NODE_NAME = 'node_name'
COL_PROC_ID = 'proc_id'
COL_PROC_NAME = 'proc_name'
COL_RANK = 'rank'
COL_SIZE = 'size'
COL_TIME = 'time'
COL_TIME_OVERALL = 'time_overall'
COL_TIME_RANGE = 'time_range'
COL_TIME_START = 'time_start'
COL_TIME_END = 'time_end'


LOGICAL_VIEW_TYPES = [
    ('file_name', 'file_dir'),
    ('file_name', 'file_pattern'),
    ('proc_name', 'app_name'),
    ('proc_name', 'host_name'),
    ('proc_name', 'node_name'),
    ('proc_name', 'proc_id'),
    ('proc_name', 'rank'),
    ('proc_name', 'thread_id'),
]
VIEW_TYPES = [
    'file_name',
    'proc_name',
    'time_range',
]

ACC_PAT_SUFFIXES = ['time', 'size', 'count']
DERIVED_MD_OPS = ['close', 'open', 'seek', 'stat']
IO_CATS = [io_cat.value for io_cat in list(IOCategory)]
IO_TYPES = ['read', 'write', 'metadata']
COMPACT_IO_TYPES = ['R', 'W', 'M']


# todo(izzet): add mmap
POSIX_IO_CAT_FUNCTIONS = {
    IOCategory.READ: {
        'fread',
        'pread',
        'preadv',
        'read',
        'readv',
    },
    IOCategory.WRITE: {
        'fwrite',
        'pwrite',
        'pwritev',
        'write',
        'writev',
    },
    IOCategory.SYNC: {
        'fsync',  # Forces write of all dirty pages for a file to disk
        'fdatasync',  # Similar to fsync but doesn't flush metadata
        'msync',  # Memory-mapped file sync
        'sync',  # System-wide sync
    },
    IOCategory.METADATA: {
        '__fxstat',
        '__fxstat64',
        '__lxstat',
        '__lxstat64',
        '__xstat',
        '__xstat64',
        'access',
        'close',
        'closedir',
        'fclose',
        'fcntl',
        'fopen',
        'fopen64',
        'fseek',
        'fstat',
        'fstatat',
        'ftell',
        'ftruncate',
        'link',
        'lseek',
        'lseek64',
        'mkdir',
        'open',
        'open64',
        'opendir',
        'readdir',
        'readlink',
        'remove',
        'rename',
        'rmdir',
        'seek',
        'stat',
        'unlink',
    },
    IOCategory.PCTL: {
        'exec',
        'exit',
        'fork',
        'kill',
        'pipe',
        'wait',
    },
    IOCategory.IPC: {
        'msgctl',
        'msgget',
        'msgrcv',
        'msgsnd',
        'semctl',
        'semget',
        'semop',
        'shmat',
        'shmctl',
        'shmdt',
        'shmget',
    },
}
POSIX_IO_CAT_MAPPING = {func: category for category, functions in POSIX_IO_CAT_FUNCTIONS.items() for func in functions}
POSIX_METADATA_FUNCTIONS = POSIX_IO_CAT_FUNCTIONS[IOCategory.METADATA]


FILE_PATTERN_PLACEHOLDER = '[0-9]'
PROC_NAME_SEPARATOR = '#'

HUMANIZED_COLS = dict(
    acc_pat='Access Pattern',
    app_io_time='Application I/O Time',
    app_name='Application',
    behavior='Behavior',
    cat='Category',
    checkpoint_io_time='Checkpoint I/O Time',
    compute_time='Compute Time',
    count='Count',
    file_dir='File Directory',
    file_name='File',
    file_pattern='File Pattern',
    func_name='Function Name',
    host_name='Host',
    io_cat='I/O Category',
    io_time='I/O Time',
    node_name='Node',
    proc_name='Process',
    rank='Rank',
    read_io_time='Read I/O Time',
    size='Size',
    time='Time',
    time_range='Time Period',
    u_app_compute_time='Unoverlapped Application Compute Time',
    u_app_io_time='Unoverlapped Application I/O Time',
    u_checkpoint_io_time='Unoverlapped Checkpoint I/O Time',
    u_compute_time='Unoverlapped Compute Time',
    u_io_time='Unoverlapped I/O Time',
    u_read_io_time='Unoverlapped Read I/O Time',
)
HUMANIZED_METRICS = dict(
    bw='I/O Bandwidth',
    intensity='I/O Intensity',
    iops='I/O Operations per Second',
    time='I/O Time',
)
HUMANIZED_VIEW_TYPES = dict(
    app_name='App',
    file_dir='File Directory',
    file_name='File',
    file_pattern='File Pattern',
    node_name='Node',
    proc_name='Process',
    rank='Rank',
    time_range='Time Period',
)

KiB = 1024.0
MiB = KiB * KiB
GiB = KiB * MiB

SIZE_BINS = [
    -np.inf,
    4 * KiB,
    16 * KiB,
    64 * KiB,
    256 * KiB,
    1 * MiB,
    4 * MiB,
    16 * MiB,
    64 * MiB,
    256 * MiB,
    1 * GiB,
    4 * GiB,
    np.inf,
]
SIZE_BIN_LABELS = [
    '<4 KiB',
    '4 KiB - 16 KiB',
    '16 KiB - 64 KiB',
    '64 KiB - 256 KiB',
    '256 KiB - 1 MiB',
    '1 MiB - 4 MiB',
    '4 MiB - 16 MiB',
    '16 MiB - 64 MiB',
    '64 MiB - 256 MiB',
    '256 MiB - 1 GiB',
    '1 GiB - 4 GiB',
    '>4 GiB',
]
SIZE_BIN_NAMES = [
    '<4 kiB',
    "4 KiB",
    "16 KiB",
    "64 KiB",
    "256 KiB",
    "1 MiB",
    "4 MiB",
    "16 MiB",
    "64 MiB",
    "256 MiB",
    "1 GiB",
    ">4 GiB",
]
SIZE_BIN_SUFFIXES = [
    "0_4kib",
    "4kib_16kib",
    "16kib_64kib",
    "64kib_256kib",
    "256kib_1mib",
    "1mib_4mib",
    "4mib_16mib",
    "16mib_64mib",
    "64mib_256mib",
    "256mib_1gib",
    "1gib_4gib",
    "4gib_plus",
]

NUM_SIZE_BINS = len(SIZE_BINS) - 1
assert len(SIZE_BIN_LABELS) == NUM_SIZE_BINS, (
    f"SIZE_BIN_LABELS length mismatch: expected {NUM_SIZE_BINS}, got {len(SIZE_BIN_LABELS)}"
)
assert len(SIZE_BIN_NAMES) == NUM_SIZE_BINS, (
    f"SIZE_BIN_NAMES length mismatch: expected {NUM_SIZE_BINS}, got {len(SIZE_BIN_NAMES)}"
)
assert len(SIZE_BIN_SUFFIXES) == NUM_SIZE_BINS, (
    f"SIZE_BIN_SUFFIXES length mismatch: expected {NUM_SIZE_BINS}, got {len(SIZE_BIN_SUFFIXES)}"
)

EVENT_ATT_REASONS = 'attach_reasons'
EVENT_COMP_HLM = 'compute_hlm'
EVENT_COMP_MAIN_VIEW = 'compute_main_view'
EVENT_COMP_METBD = 'compute_metric_boundaries'
EVENT_COMP_PERS = 'compute_perspectives'
EVENT_COMP_ROOT_VIEWS = 'compute_root_views'
EVENT_COMP_VIEW = 'compute_view'
EVENT_DET_CHAR = 'detect_characteristics'
EVENT_READ_TRACES = 'read_traces'
EVENT_SAVE_VIEWS = 'save_views'



import dask.dataframe as dd
import json
import numpy as np
import pandas as pd
from dask.distributed import Future, get_client
from typing import List, Union

from .analyzer import Analyzer
from .constants import COL_FUNC_NAME, COL_TIME, COL_TIME_END, COL_TIME_START, IO_CATS
from .types import ViewType


CAT_POSIX = 0
CAT_MAPPING = {
    0: 'posix',
}
DROPPED_COLS = [
    'app',
    'bandwidth',
    'file_id',
    'hostname',
    'index',
    'level',
    'proc',
    'proc_id',
    'rank',
    'tend',
    'thread_id',
    'tmid',
    'tstart',
]
TRACE_COL_MAPPING = {
    'duration': COL_TIME,
    'func_id': COL_FUNC_NAME,
    'tend': COL_TIME_END,
    'tstart': COL_TIME_START,
}


class RecorderAnalyzer(Analyzer):
    def read_trace(self, trace_path, extra_columns, extra_columns_fn):
        self.global_min_max = self._load_global_min_max(trace_path=trace_path)
        traces = dd.read_parquet(trace_path).rename(columns=TRACE_COL_MAPPING)
        return traces

    def postread_trace(
        self,
        traces: dd.DataFrame,
        view_types: List[ViewType],
    ) -> dd.DataFrame:
        traces[COL_TIME] = traces[COL_TIME].astype('Float64')
        traces['acc_pat'] = traces['acc_pat'].astype('Int8')
        traces['count'] = 1
        traces['count'] = traces['count'].astype('Int64')
        traces['io_cat'] = traces['io_cat'].astype('Int8')
        time_ranges = self._compute_time_ranges(
            global_min_max=self.global_min_max,
            time_granularity=self.time_granularity,
            time_resolution=self.time_resolution,
        )
        traces = (
            traces[(traces['cat'] == CAT_POSIX) & (traces['io_cat'].isin(IO_CATS))]
            .map_partitions(self._set_time_ranges, time_ranges=time_ranges)
            .drop(columns=DROPPED_COLS, errors='ignore')
        )
        traces['cat'] = 'posix'
        traces['cat'] = traces['cat'].astype('string')
        return traces

    def get_total_event_count(self, traces: dd.DataFrame) -> int:
        return traces[(traces['cat'] == CAT_POSIX) & (traces['io_cat'].isin(IO_CATS))].reduction(len, sum).persist()

    def get_unique_host_count(self, traces: dd.DataFrame):
        return traces["hostname"].nunique()

    @staticmethod
    def _compute_time_ranges(global_min_max: dict, time_granularity: float, time_resolution: float):
        tmid_min, tmid_max = global_min_max['tmid']
        time_ranges = np.arange(tmid_min, tmid_max, int(time_granularity * time_resolution))
        return get_client().scatter(time_ranges)

    @staticmethod
    def _load_global_min_max(trace_path: str) -> dict:
        with open(f"{trace_path}/global.json") as file:
            global_min_max = json.load(file)
        return global_min_max

    @staticmethod
    def _set_time_ranges(df: pd.DataFrame, time_ranges: Union[Future, np.ndarray]):
        if isinstance(time_ranges, Future):
            time_ranges = time_ranges.result()
        return df.assign(time_range=np.digitize(df['tmid'], bins=time_ranges, right=True))



import pandas as pd
from typing import Iterable, List


def deepflatten(collection, ignore_types=(bytes, str)):
    for x in collection:
        if isinstance(x, Iterable) and not isinstance(x, ignore_types):
            yield from deepflatten(x)
        else:
            yield x


def deepmerge(source: dict, destination: dict):
    for key, value in source.items():
        if isinstance(value, dict):
            deepmerge(value, destination.setdefault(key, {}))
        else:
            destination[key] = value
    return destination


def get_every_x_intervals(values: List[int]):
    every_x = pd.Series(sorted(values)).diff().value_counts()
    every_x_values = every_x[every_x >
                             every_x.std()].index.astype(int).astype(str)
    print(values, list(every_x_values))
    return '-'.join(sorted(every_x_values))


def get_intervals(values: list):
    series = pd.Series(sorted(values, reverse=True))
    grouped = series.groupby(series.diff().fillna(1).ne(-1).cumsum())
    output = grouped.apply(lambda x: str(x.iloc[0]) if len(
        x) == 1 else str(x.iloc[-1]) + '-' + str(x.iloc[0]))
    return list(reversed(output.tolist()))


def join_with_and(values: List[str]):
    if len(values) == 1:
        return values[0]
    elif len(values) == 2:
        return ' and '.join(values)
    else:
        return ', '.join(values[:-1]) + ', and ' + values[-1]



import dask
import dask.bag as db
import dask.dataframe as dd
import glob
import json
import math
import numpy as np
import os
import pandas as pd
import portion as I
import structlog
import sys
from dftracer.utils import Indexer, Reader
from dask.distributed import wait
from typing import Callable, Dict, List, Optional

from .analyzer import Analyzer
from .constants import (
    COL_ACC_PAT,
    COL_COUNT,
    COL_FILE_NAME,
    COL_FUNC_NAME,
    COL_HOST_NAME,
    COL_IO_CAT,
    COL_PROC_NAME,
    COL_TIME,
    COL_TIME_END,
    COL_TIME_RANGE,
    COL_TIME_START,
    POSIX_IO_CAT_MAPPING,
    POSIX_METADATA_FUNCTIONS,
    IOCategory,
)
from .types import ViewType
from .utils.log_utils import log_block

logger = structlog.get_logger()

CAT_POSIX = "POSIX"
CAT_STDIO = "STDIO"
COND_CHECKPOINT = {
    "cat": {"checkpoint"},
    "name": {"TFCheckpointing.checkpoint"},
}
COND_COMPUTE = {
    "cat": {"compute"},
    "name": {"TFFramework.compute", "compute", "cpu"},
}
COND_READ = {
    "cat": {"IO"},
    "name": {
        "TFReader._parse_image",
        "TorchDataset.__getitem__",
    },
}
IGNORED_FILE_PATTERNS = [
    "/dev/",
    "/etc/",
    "/gapps/python",
    "/lib/python",
    "/proc/",
    "/software/",
    "/sys/",
    "/usr/lib",
    "/usr/tce/backend",
    "/usr/tce/packages",
    "/venv",
    "__pycache__",
]
IGNORED_FUNC_NAMES = [
    "DLIOBenchmark.__init__",
    # 'DLIOBenchmark._train',
    "DLIOBenchmark.initialize",
    # 'DLIOBenchmark.run',
    "FileStorage.__init__",
    "IndexedBinaryMMapReader.__init__",
    "IndexedBinaryMMapReader.load_index",
    "IndexedBinaryMMapReader.next",
    "IndexedBinaryMMapReader.read_index",
    "NPZReader.__init__",
    "NPZReader.next",
    "NPZReader.read_index",
    "PyTorchCheckpointing.__init__",
    "PyTorchCheckpointing.finalize",
    "PyTorchCheckpointing.get_tensor",
    "SCRPyTorchCheckpointing.__init__",
    "SCRPyTorchCheckpointing.finalize",
    "SCRPyTorchCheckpointing.get_tensor",
    "TFCheckpointing.__init__",
    "TFCheckpointing.finalize",
    "TFCheckpointing.get_tensor",
    "TFDataLoader.__init__",
    "TFDataLoader.finalize",
    "TFDataLoader.next",
    "TFDataLoader.read",
    "TFFramework.get_loader",
    "TFFramework.init_loader",
    "TFFramework.is_nativeio_available",
    "TFFramework.trace_object",
    "TFReader.__init__",
    "TFReader.next",
    "TFReader.read_index",
    "TorchDataLoader.__init__",
    "TorchDataLoader.finalize",
    "TorchDataLoader.next",
    "TorchDataLoader.read",
    "TorchDataset.__init__",
    # 'TorchDataset.worker_init',
    "TorchFramework.get_loader",
    "TorchFramework.init_loader",
    "TorchFramework.is_nativeio_available",
    "TorchFramework.trace_object",
]
IGNORED_FUNC_PATTERNS = [
    '.save_state',
    'checkpoint_end_',
    'checkpoint_start_',
]
TRACE_COL_MAPPING = {
    'dur': COL_TIME,
    'name': COL_FUNC_NAME,
    'te': COL_TIME_END,
    'trange': COL_TIME_RANGE,
    'ts': COL_TIME_START,
}


def create_index(filename):
    index_file = f"{filename}.idx"
    if not os.path.exists(index_file):
        indexer = Indexer(filename, index_file, checkpoint_size=32 * 1024 * 1024)
        indexer.build()
        logger.debug("Creating index", filename=filename)
    return filename


def generate_batches(filename, max_bytes):
    batch_size = 4 * 1024 * 1024  # 4 MB
    for start in range(0, max_bytes, batch_size):
        # this range is intended since DFTracerJsonLinesBytesReader do
        # line boundary algorithm internally to chop incomplete line
        end = min(start + batch_size, max_bytes)
        logger.debug("Created batch", filename=filename, start=start, end=end)
        yield filename, start, end


def get_size(filename):
    size = 0
    if filename.endswith(".pfw"):
        size = os.stat(filename).st_size
    elif filename.endswith(".pfw.gz"):
        index_file = f"{filename}.idx"
        indexer = Indexer(filename, index_file)
        size = indexer.get_max_bytes()
    logger.debug("File has size", filename=filename, size=size / 1024**3)
    return filename, int(size)


def get_io_cat(func_name: str):
    if func_name in POSIX_METADATA_FUNCTIONS:
        return IOCategory.METADATA.value
    if func_name in POSIX_IO_CAT_MAPPING:
        return POSIX_IO_CAT_MAPPING[func_name].value
    return IOCategory.OTHER.value


def io_columns():
    columns = {
        "fhash": "string[pyarrow]",
        "hhash": "string[pyarrow]",
        "image_id": "uint64[pyarrow]",
        "io_cat": "uint8[pyarrow]",
        "size": "uint64[pyarrow]",
        "offset": "uint64[pyarrow]",
    }
    return columns


def io_function(json_dict: dict):
    d = {}
    d[COL_IO_CAT] = IOCategory.OTHER.value
    if "args" in json_dict:
        if "fhash" in json_dict["args"]:
            d["fhash"] = str(json_dict["args"]["fhash"])
        if "size_sum" in json_dict["args"]:
            d["size"] = int(json_dict["args"]["size_sum"])
        elif json_dict["cat"] in [CAT_POSIX, CAT_STDIO]:
            name = json_dict["name"]
            io_cat = get_io_cat(name)
            if "ret" in json_dict["args"]:
                size = int(json_dict["args"]["ret"])
                if size > 0:
                    if io_cat in [IOCategory.READ.value, IOCategory.WRITE.value]:
                        d["size"] = size
            if "offset" in json_dict["args"]:
                offset = int(json_dict["args"]["offset"])
                if offset >= 0:
                    d["offset"] = offset
            d[COL_IO_CAT] = io_cat
        else:
            if "image_idx" in json_dict["args"]:
                image_id = int(json_dict["args"]["image_idx"])
                if image_id > 0:
                    d["image_id"] = image_id
            # if "image_size" in json_object["args"]:
            #     name = json_object["name"].lower()
            #     # e.g. NPZReader.open image_size is not correct
            #     if 'reader.open' not in name:
            #         size = int(json_object["args"]["image_size"])
            #         if size > 0:
            #             d["size"] = size
    return d


def is_pyarrow_dtype_supported() -> bool:
    return sys.version_info >= (3, 9)


def load_indexed_gzip_files(filename, start, end):
    index_file = f"{filename}.idx"
    reader = Reader(filename, index_file)
    json_lines = reader.read_line_bytes_json(start, end)
    logger.debug("Read json lines", filename=filename, start=start, end=end, num_lines=len(json_lines))
    return json_lines


def load_objects_dict(
    json_dict: dict,
    time_approximate: bool,
    extra_columns: Optional[Dict[str, str]],
    extra_columns_fn: Optional[Callable[[dict], dict]],
):
    final_dict = {}
    logger.debug("Loading dict", json_dict=json_dict)
    if json_dict is not None:
        try:
            if "name" in json_dict:
                final_dict["name"] = json_dict["name"]
            if "cat" in json_dict:
                final_dict["cat"] = json_dict["cat"].lower()
            if "pid" in json_dict:
                final_dict["pid"] = json_dict["pid"]
            if "tid" in json_dict:
                final_dict["tid"] = json_dict["tid"]
            if "args" in json_dict:
                if "hhash" in json_dict["args"]:
                    final_dict["hhash"] = str(json_dict["args"]["hhash"])
                # if "level" in val["args"]:
                #     d["level"] = int(val["args"]["level"])
                # if (
                #     "epoch" in val["args"]
                #     and val["args"]["epoch"] != "train"
                #     and val["args"]["epoch"] != "valid"
                # ):
                #     epoch = int(val["args"]["epoch"])
                #     if epoch > 0:
                #         d["epoch"] = epoch
                if "step" in json_dict["args"]:
                    step = int(json_dict["args"]["step"])
                    if step > 0:
                        final_dict["step"] = step
            if "M" == json_dict["ph"]:
                if final_dict["name"] == "FH":
                    final_dict["type"] = 1  # 1-> file hash
                    if "args" in json_dict and "name" in json_dict["args"] and "value" in json_dict["args"]:
                        final_dict["name"] = json_dict["args"]["name"]
                        final_dict["hash"] = str(json_dict["args"]["value"])
                elif final_dict["name"] == "HH":
                    final_dict["type"] = 2  # 2-> hostname hash
                    if "args" in json_dict and "name" in json_dict["args"] and "value" in json_dict["args"]:
                        final_dict["name"] = json_dict["args"]["name"]
                        final_dict["hash"] = str(json_dict["args"]["value"])
                elif final_dict["name"] == "SH":
                    final_dict["type"] = 3  # 3-> string hash
                    if "args" in json_dict and "name" in json_dict["args"] and "value" in json_dict["args"]:
                        final_dict["name"] = json_dict["args"]["name"]
                        final_dict["hash"] = str(json_dict["args"]["value"])
                elif final_dict["name"] == "PR":
                    final_dict["type"] = 5  # 5-> process metadata
                    if "args" in json_dict and "name" in json_dict["args"] and "value" in json_dict["args"]:
                        final_dict["name"] = json_dict["args"]["name"]
                        final_dict["hash"] = str(json_dict["args"]["value"])
                else:
                    final_dict["type"] = 4  # 4-> others
                    if "args" in json_dict and "name" in json_dict["args"] and "value" in json_dict["args"]:
                        final_dict["name"] = json_dict["args"]["name"]
                        final_dict["value"] = str(json_dict["args"]["value"])
            else:
                final_dict["type"] = 0  # 0->regular event
                if "dur" in json_dict:
                    if type(json_dict["dur"]) is not int:
                        json_dict["dur"] = int(json_dict["dur"])
                    if type(json_dict["ts"]) is not int:
                        json_dict["ts"] = int(json_dict["ts"])
                    final_dict["ts"] = json_dict["ts"]
                    final_dict["dur"] = json_dict["dur"]
                    final_dict["te"] = final_dict["ts"] + final_dict["dur"]
                    if not time_approximate:
                        final_dict["tinterval"] = I.to_string(
                            I.closed(json_dict["ts"], json_dict["ts"] + json_dict["dur"])
                        )
                final_dict.update(io_function(json_dict))
                final_dict.update(extra_columns_fn(json_dict) if extra_columns_fn else {})
            # check if all extra columns are present
            if extra_columns and not all(col in final_dict for col in extra_columns):
                missing_cols = [col for col in extra_columns if col not in final_dict]
                raise ValueError(f"Missing extra columns: {missing_cols}")
            logger.debug("Built a dictionary for dict", final_dict=final_dict)
            yield final_dict
        except ValueError as error:
            logger.error("Processing dict failed", dict=json_dict, error=error)
    return {}


def load_objects_str(
    line: str,
    time_approximate: bool,
    extra_columns: Optional[Dict[str, str]],
    extra_columns_fn: Optional[Callable[[dict], dict]],
):
    if line is not None and line != "" and len(line) > 0 and "[" != line[0] and "]" != line[0] and line != "\n":
        try:
            unicode_line = "".join([i if ord(i) < 128 else "#" for i in line])
            json_dict = json.loads(unicode_line, strict=False)
            yield from load_objects_dict(json_dict, time_approximate, extra_columns, extra_columns_fn)
        except ValueError as error:
            logger.error("Processing line failed", line=line, error=error)
    return {}


class DFTracerAnalyzer(Analyzer):
    def read_trace(self, trace_path, extra_columns, extra_columns_fn):
        with log_block("glob_files"):
            pfw_pattern, pfw_gz_pattern = [], []
            if os.path.isdir(trace_path):
                pfw_pattern = glob.glob(os.path.join(trace_path, "*.pfw"))
                pfw_gz_pattern = glob.glob(os.path.join(trace_path, "*.pfw.gz"))
            elif trace_path.endswith(".pfw.gz"):
                pfw_gz_pattern = glob.glob(trace_path) if "*" in trace_path else [trace_path]
            elif trace_path.endswith(".pfw"):
                pfw_pattern = glob.glob(trace_path) if "*" in trace_path else [trace_path]
            all_files = pfw_pattern + pfw_gz_pattern
            if not all_files:
                raise FileNotFoundError("No matching .pfw or .pfw.gz files found.")
        logger.debug("Processing files", files=all_files)
        if len(pfw_gz_pattern) > 0:
            with log_block("create_index"):
                db.from_sequence(pfw_gz_pattern).map(create_index).compute()
                logger.info("Created index for files", num_files=len(pfw_gz_pattern))
        with log_block("sum_total_size"):
            sizes = db.from_sequence(all_files).map(get_size).compute()
            total_size = sum(size for _, size in sizes)
            logger.info("Total size of all files", total_size=total_size)
        gz_bag = None
        pfw_bag = None
        if len(pfw_gz_pattern) > 0:
            with log_block("gzip_index_and_batches"):
                logger.debug("Max bytes per file", sizes=sizes)
                json_line_delayed = []
                total_lines = 0
                for filename, max_bytes in sizes:
                    total_lines += max_bytes
                    for _, start, end in generate_batches(filename, max_bytes):
                        json_line_delayed.append((filename, start, end))

                logger.info(
                    "Loading batches",
                    num_batches=len(json_line_delayed),
                    num_files=len(pfw_gz_pattern),
                    total_lines=total_lines,
                )
                json_line_bags = []
                for filename, start, end in json_line_delayed:
                    json_line_bags.append(dask.delayed(load_indexed_gzip_files)(filename, start, end))
                json_lines = db.concat(json_line_bags)
            with log_block("parse_gzip_json_lines"):
                gz_bag = (
                    json_lines.map(
                        load_objects_dict,
                        time_approximate=self.time_approximate,
                        extra_columns=extra_columns,
                        extra_columns_fn=extra_columns_fn,
                    )
                    .flatten()
                    .filter(lambda x: "name" in x)
                )
        main_bag = None
        if len(pfw_pattern) > 0:
            with log_block("parse_json_lines"):
                pfw_bag = (
                    db.read_text(pfw_pattern)
                    .map(
                        load_objects_str,
                        time_approximate=self.time_approximate,
                        extra_columns=extra_columns,
                        extra_columns_fn=extra_columns_fn,
                    )
                    .flatten()
                    .filter(lambda x: "name" in x)
                )
        if len(pfw_gz_pattern) > 0 and len(pfw_pattern) > 0:
            main_bag = db.concat([pfw_bag, gz_bag])
        elif len(pfw_gz_pattern) > 0:
            main_bag = gz_bag
        elif len(pfw_pattern) > 0:
            main_bag = pfw_bag
        if main_bag:
            columns = {
                "name": "string",
                "cat": "string",
                "type": "Int8",
                "pid": "Int64",
                "tid": "Int64",
                "ts": "Int64",
                "te": "Int64",
                "dur": "Int64",
                "tinterval": "Int64" if self.time_approximate else "string",
                "trange": "Int64",
                "level": "Int8",
            }
            if is_pyarrow_dtype_supported():
                columns = {
                    "name": "string[pyarrow]",
                    "cat": "string[pyarrow]",
                    "type": "uint8[pyarrow]",
                    "pid": "uint64[pyarrow]",
                    "tid": "uint64[pyarrow]",
                    "ts": "uint64[pyarrow]",
                    "te": "uint64[pyarrow]",
                    "dur": "uint64[pyarrow]",
                    "tinterval": "uint64[pyarrow]",
                    "trange": "uint64[pyarrow]",
                    "level": "uint8[pyarrow]",
                }
                if self.time_approximate:
                    columns["tinterval"] = "string[pyarrow]"
            columns.update(io_columns())
            file_hash_columns = {
                "name": "string",
                "hash": "string",
                "pid": "Int64",
                "tid": "Int64",
                "hhash": "string",
            }
            hostname_hash_columns = {
                "name": "string",
                "hash": "string",
                "pid": "Int64",
                "tid": "Int64",
                "hhash": "string",
            }
            string_hash_columns = {
                "name": "string",
                "hash": "string",
                "pid": "Int64",
                "tid": "Int64",
                "hhash": "string",
            }
            other_metadata_columns = {
                "name": "string",
                "value": "string",
                "pid": "Int64",
                "tid": "Int64",
                "hhash": "string",
            }
            if is_pyarrow_dtype_supported():
                file_hash_columns = {
                    'name': "string[pyarrow]",
                    'hash': "string[pyarrow]",
                    'pid': "uint64[pyarrow]",
                    'tid': "uint64[pyarrow]",
                    'hhash': "string[pyarrow]",
                }
                hostname_hash_columns = {
                    'name': "string[pyarrow]",
                    'hash': "string[pyarrow]",
                    'pid': "uint64[pyarrow]",
                    'tid': "uint64[pyarrow]",
                    'hhash': "string[pyarrow]",
                }
                string_hash_columns = {
                    'name': "string[pyarrow]",
                    'hash': "string[pyarrow]",
                    'pid': "uint64[pyarrow]",
                    'tid': "uint64[pyarrow]",
                    'hhash': "string[pyarrow]",
                }
                other_metadata_columns = {
                    'name': "string[pyarrow]",
                    'value': "string[pyarrow]",
                    'pid': "uint64[pyarrow]",
                    'tid': "uint64[pyarrow]",
                    'hhash': "string[pyarrow]",
                }
            columns.update(file_hash_columns)
            columns.update(hostname_hash_columns)
            columns.update(string_hash_columns)
            columns.update(other_metadata_columns)
            columns.update(extra_columns or {})
            with log_block("to_dataframe+filter_events"):
                self.all_events = main_bag.to_dataframe(meta=columns)
                events = self.all_events.query("type == 0")
            self.file_hash = (
                self.all_events.query("type == 1")[list(file_hash_columns.keys())].groupby("hash").first().persist()
            )
            self.host_hash = (
                self.all_events.query("type == 2")[list(hostname_hash_columns.keys())]
                .groupby("hash")
                .first()
                .persist()
            )
            self.string_hash = (
                self.all_events.query("type == 3")[list(string_hash_columns.keys())].groupby("hash").first().persist()
            )
            self.metadata = self.all_events.query("type == 4")[list(other_metadata_columns.keys())].persist()
            with log_block("repartition+persist+indexing"):
                self.n_partition = math.ceil(total_size / (128 * 1024**2))
                logger.debug("Number of partitions used", n_partition=self.n_partition)
                self.events = events.repartition(npartitions=self.n_partition).persist()
                _ = wait(self.events)
                self.events["ts"] = self.events["ts"] - self.events["ts"].min()
                self.events["te"] = self.events["ts"] + self.events["dur"]
                self.events["trange"] = self.events["ts"] // (self.time_granularity * self.time_resolution)
                if is_pyarrow_dtype_supported():
                    self.events["ts"] = self.events["ts"].astype("uint64[pyarrow]")
                    self.events["te"] = self.events["te"].astype("uint64[pyarrow]")
                    self.events["trange"] = self.events["trange"].astype("uint16[pyarrow]")
                else:
                    self.events["ts"] = self.events["ts"].astype("Int64")
                    self.events["te"] = self.events["te"].astype("Int64")
                    self.events["trange"] = self.events["trange"].astype("Int16")
                self.events = self.events.persist()
                _ = wait(
                    [
                        self.file_hash,
                        self.host_hash,
                        self.string_hash,
                        self.metadata,
                        self.events,
                    ]
                )
        else:
            logger.error("Unable to load traces")
            exit(1)
        # ===============================================
        with log_block("normalize_durations"):
            self.events["dur"] = self.events["dur"] / self.time_resolution

        file_hashes = self.file_hash[["name"]].rename(columns={"name": COL_FILE_NAME})
        host_hhash_empty = self.host_hash["hhash"].isna().all().compute()
        if host_hhash_empty:
            host_hashes = self.host_hash[["name"]].rename(columns={"name": COL_HOST_NAME})
        else:
            host_hashes = self.host_hash.set_index("hhash")[["name"]].rename(columns={"name": COL_HOST_NAME})

        with log_block("join_hashes"):
            self.events = (
                self.events.merge(
                    file_hashes,
                    how="left",
                    left_on="fhash",
                    right_index=True,
                )
                .merge(
                    host_hashes,
                    how="left",
                    left_on="hhash",
                    right_index=True,
                )
                .drop(columns=["fhash", "hhash"])
            )

        with log_block("finalize+rename_columns+persist"):
            self.events = self.events.rename(columns=TRACE_COL_MAPPING).persist()
            _ = wait(self.events)
            return self.events

    def postread_trace(
        self,
        traces: dd.DataFrame,
        view_types: List[ViewType],
    ) -> dd.DataFrame:
        with log_block("filter_files"):
            traces = traces[
                traces[COL_FILE_NAME].isna() | ~traces[COL_FILE_NAME].str.contains("|".join(IGNORED_FILE_PATTERNS))
            ]

        # Ignore redundant function calls
        with log_block("filter_functions"):
            traces = traces[~traces[COL_FUNC_NAME].isin(IGNORED_FUNC_NAMES)]
            traces = traces[~traces[COL_FUNC_NAME].str.contains("|".join(IGNORED_FUNC_PATTERNS))]

        with log_block("wait"):
            _ = wait(traces)

        with log_block("set_basic_columns"):
            traces[COL_ACC_PAT] = 0
            traces[COL_COUNT] = 1

        # drop columns that are not needed
        # if COL_FILE_NAME not in view_types:
        #     traces = traces.drop(columns=[COL_FILE_NAME], errors='ignore')
        # if COL_HOST_NAME not in view_types:
        #     traces = traces.drop(columns=[COL_HOST_NAME], errors='ignore')

        # Set batches
        # traces['batch'] = traces.groupby(['func_name', 'step']).cumcount() + 1
        # batch_counts = traces['batch'].value_counts()
        # last_valid_batch = batch_counts[batch_counts > 1].index.max()
        # traces['batch'] = traces['batch'].mask(
        #     traces['batch'] > last_valid_batch, pd.NA
        # )

        # pytorch reads images instead of batches
        # e.g. 4 workers = 0..4 images = who starts/finishes first

        # epoch and step make sense in dlio layer

        # to put step back, target variable = previous compute + my io

        # Set steps depending on time ranges
        # step_time_ranges = traces.groupby(['pid', 'epoch', 'step']).agg({'ts': min, 'te': max})
        # traces = traces.map_partitions(
        #     self._set_steps, step_time_ranges=step_time_ranges.reset_index()
        # )

        return (
            traces.map_partitions(self._set_proc_names)
            .map_partitions(self._fix_file_posix_category)
            .map_partitions(self._sanitize_size_offset)
        )

    def get_job_time(self, traces):
        return super().get_job_time(traces) / self.time_resolution

    def get_unique_process_count(self, traces: dd.DataFrame):
        return traces["pid"].nunique()

    @staticmethod
    def _set_epochs(df: pd.DataFrame, epochs: pd.DataFrame):
        return df.assign(epoch=np.digitize(df["time_range"], bins=epochs["time_range"], right=False))

    @staticmethod
    def _fix_file_posix_category(df: pd.DataFrame):
        base_condition = df["cat"].str.contains("posix|stdio") & ~df["file_name"].isna()

        # Step 1: Map file purpose suffixes first
        purpose_updates = {"/data": "_reader", "/checkpoint": "_checkpoint"}

        for path, suffix in purpose_updates.items():
            mask = base_condition & df["file_name"].str.contains(path)
            df.loc[mask, "cat"] = df.loc[mask, "cat"] + suffix

        # Step 2: Map filesystem suffixes
        filesystem_updates = {"/lustre": "_lustre", "/ssd": "_ssd"}

        for path, suffix in filesystem_updates.items():
            mask = base_condition & df["file_name"].str.contains(path)
            df.loc[mask, "cat"] = df.loc[mask, "cat"] + suffix

        return df

    @staticmethod
    def _sanitize_size_offset(df: pd.DataFrame):
        df["size"] = df["size"].replace(0, np.nan)
        if "offset" in df.columns:
            df["offset"] = df["offset"].replace(0, np.nan)
        return df

    @staticmethod
    def _set_proc_names(df: pd.DataFrame):
        df[COL_PROC_NAME] = (
            "app#" + df[COL_HOST_NAME].astype(str) + "#" + df["pid"].astype(str) + "#" + df["tid"].astype(str)
        )
        return df

    @staticmethod
    def _set_steps(df: pd.DataFrame, step_time_ranges: pd.DataFrame):
        mapped_traces = df.copy()

        for pid in df["pid"].unique():
            pid_trace_cond = mapped_traces["pid"] == pid
            pid_traces = mapped_traces[pid_trace_cond]
            pid_step_ranges = step_time_ranges[step_time_ranges["pid"] == pid]

            # Sort step ranges by start timestamp
            pid_step_ranges_sorted = pid_step_ranges.sort_values("ts")

            # Create bins and labels
            bins = pid_step_ranges_sorted["ts"].tolist()
            if len(bins) > 0:
                bins.append(pid_step_ranges_sorted["te"].max())
            # print(pid, bins)
            steps = pid_step_ranges_sorted["step"].tolist()

            # Use np.digitize to find bin indices
            bin_indices = np.digitize(pid_traces["ts"], bins=bins) - 1

            # Map indices to steps, leaving as None for out-of-range timestamps
            mapped_traces.loc[pid_trace_cond, "step"] = [
                steps[idx] if 0 <= idx < len(steps) else pd.NA for idx in bin_indices
            ]

        return mapped_traces



import os


def get_bool_env_var(name: str, default: bool = False) -> bool:
    return os.environ.get(name, str(default)).lower() in ["true", "1", "yes"]


def get_int_env_var(name: str, default: int = 0) -> int:
    try:
        return int(os.environ.get(name, default))
    except ValueError:
        return default



import hydra
import signal
from dask_jobqueue import LSFCluster, PBSCluster, SLURMCluster
from dataclasses import asdict, dataclass, field
from distributed import LocalCluster
from hydra.core.config_store import ConfigStore
from hydra.utils import instantiate
from omegaconf import MISSING
from typing import Any, List, Optional, Union

from .config import (
    ClusterConfig,
    CustomHelpConfig,
    CustomJobConfig,
    LocalClusterConfig,
    LSFClusterConfig,
    PBSClusterConfig,
    SLURMClusterConfig,
)


@dataclass
class ExternalCluster:
    restart_on_connect: Optional[bool]
    scheduler_address: str
    local_directory: Optional[str]


@dataclass
class Config:
    defaults: List[Any] = field(
        default_factory=lambda: [
            {"hydra/job": "custom"},
            {"cluster": "local"},
            "_self_",
            {"override hydra/help": "custom"},
        ]
    )
    cluster: ClusterConfig = MISSING
    debug: Optional[bool] = False
    verbose: Optional[bool] = False


cs = ConfigStore.instance()
cs.store(group="hydra/help", name="custom", node=asdict(CustomHelpConfig()))
cs.store(group="hydra/job", name="custom", node=CustomJobConfig)
cs.store(name="config", node=Config)
cs.store(group="cluster", name="local", node=LocalClusterConfig)
cs.store(group="cluster", name="lsf", node=LSFClusterConfig)
cs.store(group="cluster", name="pbs", node=PBSClusterConfig)
cs.store(group="cluster", name="slurm", node=SLURMClusterConfig)


ClusterType = Union[ExternalCluster, LocalCluster, LSFCluster, PBSCluster, SLURMCluster]


@hydra.main(version_base=None, config_name="config")
def main(cfg: Config) -> None:
    cluster: ClusterType = instantiate(cfg.cluster)
    print(cluster.scheduler.address, flush=True)
    try:
        signal.pause()
    except KeyboardInterrupt:
        print("Shutting down the Dask cluster...")
    finally:
        cluster.close()
        print("Dask cluster is shut down.")


if __name__ == "__main__":
    main()



import altair as alt
import dask
import shutil
import streamlit as st
import numpy as np
import pandas as pd
import tempfile
from dftracer.analyzer import init_with_hydra
from dftracer.analyzer.constants import XFER_SIZE_BIN_LABELS
from dftracer.analyzer.rules import KnownCharacteristics
from dftracer.analyzer.types import Characteristics, RawStats

DEFAULT_THRESHOLD = 45
DEFAULT_TIME_GRANULARITY_IN_SECONDS = 5  # 5 seconds
XFER_SIZE_CAT_TYPE = pd.CategoricalDtype(categories=XFER_SIZE_BIN_LABELS, ordered=True)
VIEW_TYPE_MAPPING = {
    'File': 'file_name',
    'Process': 'proc_name',
    'Timeline': 'time_range',
}

st.set_page_config(
    page_title="DFAnalyzer Live",
    layout="centered",
    menu_items={
        'About': 'https://grc.iit.edu/research/projects/wisio',
        'Report a bug': 'https://github.com/grc-iit/wisio/issues',
    },
)

st.write(
    r'''
    <style>
        [data-testid="stImageContainer"] img {border-radius: 0;}
        [data-testid="stMainBlockContainer"] {max-width: 812px;}
    </style>
    ''',
    unsafe_allow_html=True,
)

st.image("assets/logo.png", width=200)
st.title("Welcome to WisIO Live")
st.markdown(
    """
    Analyze, visualize, and understand I/O performance issues in HPC workloads.
    """
)

result = None
bottlenecks = None
characteristics: Characteristics = {}
raw_stats: RawStats = {}

with st.form('analysis_form'):
    trace_files = st.file_uploader(
        "Upload a trace file",
        type=["darshan", "parquet", "pfw", "pfw.gz"],
        accept_multiple_files=True,
    )

    view_types = st.multiselect(
        "Select perspectives to analyze",
        options=VIEW_TYPE_MAPPING.keys(),
        default=VIEW_TYPE_MAPPING.keys(),
    )

    time_granularity = st.slider(
        "Set time granularity for analysis (in seconds)",
        min_value=1,
        max_value=100,
        value=DEFAULT_TIME_GRANULARITY_IN_SECONDS,
        step=1,
        help="This sets the granularity of time intervals for analysis.",
        disabled='Timeline' not in view_types,
    )

    threshold = st.slider(
        "Set the threshold for bottleneck detection",
        min_value=0,
        max_value=90,
        format="%d%%",
        value=DEFAULT_THRESHOLD,
        step=1,
        help="This threshold determines the sensitivity of bottleneck detection.",
    )

    logical_view_types = st.checkbox(
        "Enable logical view types",
        value=False,
        help="Logical view types allow for more complex analysis but may take longer to compute.",
    )

    submit = st.form_submit_button("Analyze")

if submit:
    # Check if all trace files have the same type
    if not trace_files or len(trace_files) == 0:
        st.error("Please upload at least one trace file.")
        st.stop()
    if len(set(file.name.split('.')[-1] for file in trace_files)) > 1:
        st.error("All trace files must be of the same type.")
        st.stop()

    analyzer = 'darshan'
    if all(file.name.endswith('.parquet') for file in trace_files):
        analyzer = 'recorder'
    elif all(file.name.endswith('.pfw') or file.name.endswith('.pfw.gz') for file in trace_files):
        analyzer = 'dftracer'

    with st.status("Analyzing trace files", expanded=True) as status:
        st.write(f"Detected analyzer type: {analyzer.title()}")

        with tempfile.TemporaryDirectory() as temp_dir:
            st.write(f"Using temporary directory: {temp_dir}")

            for trace_file in trace_files:
                with open(f"{temp_dir}/{trace_file.name}", "wb") as temp_trace_file:
                    temp_trace_file.write(trace_file.getbuffer())

            wis = init_with_hydra(
                hydra_overrides=[
                    f"+analyzer={analyzer}",
                    f"analyzer.checkpoint={False}",
                    f"analyzer.time_granularity={time_granularity}",
                    f"hydra.run.dir={temp_dir}",
                    f"hydra.runtime.output_dir={temp_dir}",
                    f"logical_view_types={logical_view_types}",
                    f"threshold={threshold}",
                    f"trace_path={temp_dir}",
                    f"view_types=[{','.join([VIEW_TYPE_MAPPING[view_type] for view_type in view_types])}]",
                ]
            )
            st.write("Initialized WisIO analyzer.")

            st.write("Analyzing trace files...")
            result = wis.analyze_trace()
            (characteristics, raw_stats) = dask.compute(
                result.characteristics,
                result.raw_stats,
            )
            st.write("Analysis complete.")

            try:
                st.write("Shutting down analyzer...")
                wis.client.close()
                wis.cluster.close()  # type: ignore
                st.write("Analyzer shut down.")
            except Exception as e:
                st.error(f"Error shutting down analyzer: {e}")
                st.write("Please restart the application.")

            st.write("Cleaning up temporary directory...")
            shutil.rmtree(temp_dir, ignore_errors=True)
            st.write("Temporary directory cleaned up.")

            status.update(label="Analysis complete.", expanded=False, state="complete")

            st.session_state['result'] = result
            st.session_state['characteristics'] = characteristics
            st.session_state['raw_stats'] = raw_stats

# if 'result' in st.session_state:
#     result = st.session_state['result']
#     bottlenecks = st.session_state['bottlenecks']
#     characteristics = st.session_state['characteristics']
#     raw_stats = st.session_state['raw_stats']
# else:
#     result = None

if result:
    st.subheader("Analysis Results")

    characteristics_tab, bottlenecks_tab = st.tabs(["I/O Characteristics", "I/O Bottlenecks"])

    with characteristics_tab:
        file_count = characteristics[KnownCharacteristics.FILE_COUNT.value].value
        proc_count = characteristics[KnownCharacteristics.PROC_COUNT.value].value
        io_ops = characteristics[KnownCharacteristics.IO_COUNT.value].value
        io_size_fmt = characteristics[KnownCharacteristics.IO_SIZE.value].value_fmt
        io_time = characteristics[KnownCharacteristics.IO_TIME.value].value
        read_xfer_bins = characteristics[KnownCharacteristics.READ_XFER_SIZE.value]._dataframe
        write_xfer_bins = characteristics[KnownCharacteristics.WRITE_XFER_SIZE.value]._dataframe

        col11, col12, col13 = st.columns(3)
        col11.metric("Runtime", f"{raw_stats.job_time:.2f} s", border=True)
        col12.metric(r"\# of Processes", f"{file_count:,}", border=True)
        col13.metric(r"\# of Files", f"{proc_count:,}", border=True)

        col21, col22, col23 = st.columns(3)
        col21.metric("I/O Time", f"{io_time:.2f} s", border=True)
        col22.metric("I/O Operations", f"{io_ops:,}", border=True)
        col23.metric("I/O Size", io_size_fmt, border=True)

        col31, col32 = st.columns(2)
        col31.markdown("**Read Request Size Distribution**")
        read_xfer_bins_full = read_xfer_bins['read_count'].reindex(XFER_SIZE_BIN_LABELS).fillna(0)
        read_xfer_bins_fixed = pd.DataFrame(
            {"Size Range": read_xfer_bins_full.index, "Operations": read_xfer_bins_full.values}
        )
        read_xfer_bins_fixed['Size Range'] = read_xfer_bins_fixed['Size Range'].astype(XFER_SIZE_CAT_TYPE)
        col31.write(
            alt.Chart(read_xfer_bins_fixed)
            .mark_bar()
            .encode(
                x=alt.X('Operations', title='# of I/O Operations'),
                y=alt.Y('Size Range', sort=None, title=None),
            )
        )
        # col31.bar_chart(read_xfer_bins_fixed.set_index('Size Range'), horizontal=True)
        col32.markdown("**Write Request Size Distribution**")
        write_xfer_bins_fixed = write_xfer_bins['write_count'].reindex(XFER_SIZE_BIN_LABELS).fillna(0)
        write_xfer_bins_fixed = pd.DataFrame(
            {"Size Range": write_xfer_bins_fixed.index, "Operations": write_xfer_bins_fixed.values}
        )
        write_xfer_bins_fixed['Size Range'] = write_xfer_bins_fixed['Size Range'].astype(XFER_SIZE_CAT_TYPE)
        col32.write(
            alt.Chart(write_xfer_bins_fixed)
            .mark_bar()
            .encode(
                x=alt.X('Operations', title='# of I/O Operations'),
                y=alt.Y('Size Range', sort=None, title=None),
            )
        )

    with bottlenecks_tab:
        st.write(bottlenecks)

        # st.subheader("Time View (4 bottlenecks with 7 reasons)")

        # with st.expander("[CR1] 32 processes, 2 files, I/O Time: 2.19s (53.26%)"):
        #     st.markdown("""
        #     - **[Excessive metadata access]** Overall **100.00%** (2.19 seconds) of I/O time is spent on metadata access.
        #         - Specifically, **100.00%** (2.19 seconds) is on the 'open' operation.
        #     """)

        # with st.expander("[CR2] 1 process, 6 files, I/O Time: 0.33s (7.97%)"):
        #     st.markdown("""
        #     - **[Excessive metadata access]** Overall **99.35%** (0.33 seconds) of I/O time is spent on metadata access.
        #         - Specifically, **99.13%** (0.33 seconds) is on the 'open' operation.
        #     """)
